// nn.c Copyright 1995-98 by Donald R. Tveter

#include <stdio.h>
#include <math.h>
#include <signal.h>
#include <setjmp.h>

#ifdef BSD
#include <termios.h>
#elif defined NEXT
#include <malloc.h>
#include <sgtty.h>
#elif defined UNIX // SysV, Linux
#include <malloc.h>
#include <termio.h>
#endif

int pg(char *);

#if defined(UNIX) && defined(HOTKEYS)

#include <sys/ioctl.h>
#include <unistd.h>
#include <stdlib.h>

#ifdef BSD
struct termios raw;    // characteristics for hotkeys 
struct termios noraw;  // original characteristics 
#elif defined NEXT
struct sgttyb raw;     // characteristics for hotkeys 
struct sgttyb noraw;   // original characteristics 
#else // the System V solution 
struct termio raw;    // characteristics for hotkeys 
struct termio noraw;  // original characteristics 
#endif

char onechar[1];      // the buffer for the hotkey character 
#endif

#ifdef UNIX
#define REAL double
#define INT32 int
#define SIGINT 2
#elif defined DOS16
#include <stdlib.h>
#include <conio.h>
#define MAXINT 32767
#define INT32 long
#define REAL float
#else
#include <stdlib.h>
#include <conio.h>
#define INT32 int
#define REAL float
#define MAXINT 281474976710655
#endif

#define BUFFSIZE 258
#define GE 1
#define GT 0
#define MAXVAL 9.99999e+30
#define OUTSTRSIZE 257
#define scale(x) x
#define TRAIN 0
#define TEST 1
#define OK 0
#define FAILURE 1

typedef struct node // basic structure for each point 
   {
     int correctclass;    // the correct classification 
     int nn;              // the answer based on nn 
     REAL *cvector;       // the vector of coordinates 
     struct node *next;   // to the next point 
   } NODE;

typedef struct pat
   {
     NODE *point;          // used in finding k nearest neighbors 
     REAL distsq;
     struct pat *next;
   } PAT;

typedef struct cpat        // used in clustering patterns 
    {
      int cnt;
      struct cpat *next;
      REAL *cvector;
    } CPAT;

typedef struct data
   {
    int npats;        // number of patterns 
    int iterno;       // iteration number when data was found 
    int right;        // number of patterns right 
    int wrong;        // number of patterns wrong 
    float pctright;   // percentage right 
   } DATA;

char algorithm;       // flags lvq or dsm 
REAL alpha;           // learning rate 
NODE *blank;          // blank test pattern 
int bufferend;        // index of last character in input line 
int bufferptr;        // position of next character in buffer 
char buffer[BUFFSIZE];// holds contents of one input line 
int ch;               // general purpose character variable 
PAT *closest;         // ptr to list of nearest neighbor patterns 
CPAT *clusterstart;   // start of the list of clusters 
jmp_buf cmdloopstate; // to save state in case of a SIGINT 
FILE *copy;           // file to copy interactive session to 
char copyflag;        // flags the interactive copy 
int count;            // number of clusters in all 
int create;           // number of clusters created in ONE pass 
FILE *dataf;          // file for original data 
char *datafilename;   // copy of the data file name saved here 
int destroy;          // number of clusters removed 
char echo;            // controls echoing of characters during input 
char emptystring;     // for unused string values 
INT32 *errmat;        // addr of dynamic array for error matrix 
int errmatsize;       // size of error matrix 
INT32 *errtotal;      // addr of dynamic array for error totals 
REAL fract;           // fraction of patterns to ignore when clustering 
FILE *debug;
FILE *filestack[4];   // allows for nested reads from files 
int filestackptr;     // has the index of the current file 
REAL highalpha;       // alpha to start training with 
char informat;        // controls format to read numbers 
REAL initiallambda;   // starting lambda 
char *inputfile;      // name of file to take extra commands from 
int iter;             // for counting iterations in one run 
int k;                // the k for k-nearest neighbor 
REAL lambda;          // lambda in the current pass 
int length;           // length of input vector 
int lineno;           // counts lines for pg 
int listcount;        // returns number of patterns from readpats 
char listpat;         // flags writing of patterns when evaluating them 
REAL lowalpha;        // alpha to end with 
REAL maxdist;         // distance used by the clustering routine 
int maxpasses;        // maximum number of passes thru cluster 
int nclasses;         // maximum number of possible answers 
int npats;            // number of training patterns 
REAL oneminus;        // one minus lambda 
char *original;       // original prototype file 
char outformat;       // controls format to print output 
int outofdate;        // flags out of date test stats 
char outstr[OUTSTRSIZE]; // the output string 
NODE *initialprotos;  // a copy of the initial prototype points 
char *nearestfile;    // file to write nearest neighbors to 
char *newproto;       // new prototype file 
int npasses;          // number of passes to make thru training set 
int nprotos;          // number of prototype points 
int ntestpats;        // number of test patterns 
int pagesize;         // size of page on screen 
char *protofile;      // ptr to start of string 
NODE *protolist;      // linked list of prototypes 
int readerror;        // flags an error in reading a value 
int readingpattern;   // flags reading pattern state 
int right;            // number of training patterns learned 
DATA s[2];            // statistics on training and test sets 
char summary;         // flags summary output mode 
char *testfile;       // file to take test patterns from 
NODE *testlist;       // the list of test patterns 
char *trainfile;      // file to take training patterns from 
NODE *trainlist;      // linked list of training patterns 
int unstable;         // flags a changing number of clusters 
int unlearned;        // number unlearned in last learning cycle 
INT32 *votes;         // an array to count votes per category 
int wrong;            // number of training patterns unlearned 
char wtformat;        // controls format to save and restore weights 

//#ifndef UNIX
//void interrupt(int i)
//Hitting the escape key stops the program in   
//selected loops in an emergency.  This was     
//used in debugging.  The variable, i was the   
//line number but since the lines keep shifting 
//around the numbers are no longer correct.     
//{
//if (kbhit() && getch() == 033)
//   {printf("interrupt at %d\n",i); exit(1);};
//}
//
//void hit(int i)    // another debugging function, not now in use.   
//{printf("hit line %d\n",i);}
//#endif


int pushfile(char *filename)
{
FILE *file;
file = fopen(filename,"r");
if (file == NULL)
 {
  sprintf(outstr,"cannot open:, %s\n",filename); pg(outstr);
  return(0);
 };
filestackptr = filestackptr + 1;
if (filestackptr > 3)
 {
  pg("can't stack up any more files\n");
  filestackptr = filestackptr - 1;
  fclose(file);
  return(0);
 };
filestack[filestackptr] = file;
dataf = file;
bufferptr = 0;
bufferend = 0;
buffer[0] = '\n';
buffer[1] = '\0';
return(1);
} 

#if defined(UNIX) && defined(HOTKEYS)

#ifdef BSD
void initraw()
{
 int i;
 tcgetattr(0,&noraw);
 raw.c_iflag = noraw.c_iflag;
 raw.c_oflag = noraw.c_oflag;
 raw.c_cflag = noraw.c_cflag;
 raw.c_lflag = noraw.c_lflag & ~ ECHO;
 raw.c_lflag = raw.c_lflag & ~ ICANON;
 for (i=0;i<NCCS;i++) raw.c_cc[i] = noraw.c_cc[i];
 raw.c_cc[VMIN] = 1;
 raw.c_cc[VTIME] = 0;
 raw.c_ispeed = noraw.c_ispeed;
 raw.c_ospeed = noraw.c_ospeed;
}

void setraw()
{
 tcsetattr(0,TCSANOW,&raw);
}

void setnoraw()
{
 tcsetattr(0,TCSANOW,&noraw);
}

#elif defined NEXT
void initraw()
{
 ioctl(0,TIOCGETD,&noraw);
 raw.sg_ispeed = noraw.sg_ispeed;
 raw.sg_ospeed = noraw.sg_ospeed;
 raw.sg_erase = noraw.sg_erase;
 raw.sg_kill = noraw.sg_kill;
 raw.sg_flags = noraw.sg_flags & 0000040;   // turn on raw mode 
}

void setraw()
{
 ioctl(0,TIOCSETD,&raw);
}

void setnoraw()
{
 ioctl(0,TIOCSETD,&noraw);
}

#else // System V code 

void initraw()
{
 int i;
 ioctl(0,TCGETA,&noraw);
 raw.c_iflag = noraw.c_iflag;
 raw.c_oflag = noraw.c_oflag;
 raw.c_cflag = noraw.c_cflag;
 raw.c_lflag = noraw.c_lflag & ~ 0x02;   // hot key bit 
 raw.c_lflag = raw.c_lflag & ~ 010;      // echo bit 
 raw.c_line = noraw.c_line;
 for (i=0;i<=7;i++) raw.c_cc[i] = noraw.c_cc[i];
 raw.c_cc[4] = 1;
}

void setraw()
{
 ioctl(0,TCSETA,&raw);
}

void setnoraw()
{
 ioctl(0,TCSETA,&noraw);
}
#endif // end of the System V code 

int getch()
{
 int ch;
 setraw();
 do ch = read(0,onechar,1); while (ch != 1);
 setnoraw();
 return(onechar[0]);
}
#endif  // end of the if defined(UNIX) ... code 

int readch() // returns the next character in the input buffer 
{ 
 int i, ch2;

if (bufferptr > bufferend) // then read next line into buffer 
 {
  ch2 = getc(dataf);
  if (ch2 == EOF) return(ch2);
  i = 0;
  while(ch2 != '\n' && i < (BUFFSIZE-2))
   {
#ifdef UNIX
    if (ch2 == 13) ch2 = ' ';  // turn a ctrl-M into a blank 
#endif
    if (ch2 == '\t') ch2 = ' ';
    buffer[i] = ch2;
    i = i + 1;
    ch2 = getc(dataf);
   };
  if (i == (BUFFSIZE-2)) pg("line too long\n");
  buffer[i] = '\n';
  buffer[i+1] = '\0';
  bufferend = i;
  bufferptr = 0;
  if (echo == '+') for(i = 0; i <= bufferend; i++) putchar(buffer[i]);
  if (copy && ((dataf == stdin) || (echo == '+')))
   for (i=0;i<=bufferend;i++) putc(buffer[i],copy);
 } // end if bufferptr > bufferend 
ch2 = buffer[bufferptr];
bufferptr = bufferptr + 1;
return(ch2);
}

int pg(char *str) // paging and making a copy function 
{
char *ch3,action,cr;
int i;
int nbackspaces = 5;

action = 0;
ch3 = str;
while (*ch3 != '\0')
 {
  if (*ch3 == '\n')
   {
    putchar('\n');
    if (copyflag == '+') putc('\n',copy);
    lineno = lineno + 1;
    if (pagesize > 0 && lineno > 0 && lineno % pagesize == 0)
     {
      printf("More?");
#ifdef HOTKEYS
      action = getch();
#else
      action = getchar();
      if (action != '\n') cr = getchar();
#endif
      for(i=1;i<=nbackspaces;i++) putchar(8);
      for(i=1;i<=nbackspaces;i++) putchar(' ');
      for(i=1;i<=nbackspaces;i++) putchar(8);
      if (action == 'q') goto stoppaging;
      else if (action == 'c') lineno = -MAXINT;
#ifdef UNIX
      else if (action == '\n') lineno = lineno - 1;
      else if (action == 13) lineno = lineno - 1;
#else
      else if (action == 13) lineno = lineno - 1;
#endif
      else if (action == 4) lineno = pagesize / 2;
      else {printf("action = %d\n",action);lineno = 0;};
     }
   }
  else {putchar(*ch3); if (copyflag == '+') putc(*ch3,copy); };
  ch3++;
 };
stoppaging:
if (action == 'q') return(1); else return(0);
}

int popfile()
{
ch = ' ';
if (filestackptr > 0)
 {
  fclose(dataf);
  filestackptr = filestackptr - 1;
 }
else
 {
  pg("\nunexpected EOF:  to quit the program, type q\n");
  return(FAILURE);
 };
dataf = filestack[filestackptr];
bufferptr = 0;
bufferend = 0;
buffer[0] = '\n';
buffer[1] = '\0';
return(OK);
}

void texterror(int location)
{
int ch2;
pg("unexpected text:  ");
bufferptr = bufferptr - 1;
do {ch2 = readch(); putchar(ch2);} while (ch2 != '\n');
pg("\n");
sprintf(outstr,"from %d\n",location); pg(outstr);
bufferptr = bufferptr - 1;
}

char *readstr()
{
short i,start,end;
char *addr, *addr2;
i = bufferptr;
while (buffer[i] == ' ') i = i + 1;
start = i;
while (buffer[i] != ' ' && buffer[i] != '\n') i = i + 1;
end = i-1;
addr = (char *) malloc((int) end-start+2);
if (addr == NULL)
   {pg("ran out of memory\n"); exit(420);};
addr2 = addr;
for (i=start;i<=end;i++) *addr++ = buffer[i];
bufferptr = end + 1;
*addr = '\0';
return(addr2);
}

int scanfordigit()
{
int sign, ch2;

sign = 1;

restart:
do ch2 = readch(); while (ch2 == ' ' || ch2 == '\n');
if ((ch2 >= '0' && ch2 <= '9') || ch2 == '.')
 {
  bufferptr = bufferptr - 1;
  return(sign);
 };
          switch (ch2) {
case EOF: readerror = 2;
          return(0);
case '*': while (ch2 != '\n') ch2 = readch();
          goto restart;
case '-': sign = -sign;
          goto restart;
default:  readerror = 1;
          return(0);
          };
}

int readint(int min,int max,char command)
{
int sign, number, ch2;

readerror = 0;
sign = scanfordigit();
if (readerror)
 {
  if (readerror == 1) texterror(477);
  return(0);
 };
number = 0;
do ch2 = readch(); while (ch2 == ' ');
while (ch2 >= '0' && ch2 <= '9')
 {
  number = number * 10 + (ch2 - '0');
  ch2 = readch();
 };
bufferptr = bufferptr - 1;
number = sign * number;
if (number < min || number > max)
 {
  sprintf(outstr,"out of range value: %d",number); pg(outstr);
  if (dataf == stdin) pg("\n");
  else
   {
    sprintf(outstr," in %c command\n",command);
    pg(outstr);
   };
  readerror = 1;
 };
return(number);
}

REAL readreal(int op,REAL min,int command)
{
REAL fractpart, divisor, intpart, sign;
double number;
int ch2, numread, i;

readerror = 0;
sign = (REAL) scanfordigit();
if (readerror || (sign == 0 && !readingpattern))
 {
  if (readerror == 1) texterror(516);
  return(0);
 };
ch2 = readch();
i = 0;
while ((ch2 >= '0' && ch2 <='9') || ch2 == '+' || ch2 == '-' ||
       ch2 == 'e' || ch2 == 'E' || ch2 == '.')
 {
  outstr[i] = ch2;
  i = i+1;
  ch2 = readch();
 };
outstr[i] = 0;
bufferptr = bufferptr - 1;
numread = sscanf(outstr,"%lf",&number);
if (numread == 0)
 {
  pg("bad real value\n");
  readerror = 1;
  return(0.0);
 };
number = sign * number;
if (op == GT && number > min) return(number);
else if (op == GE && number >= min) return(number);
else
 {
  sprintf(outstr,"erroneous value: %f",number); pg(outstr);
  if (dataf == stdin) pg("\n");
  else {sprintf(outstr," in %c command\n",command); pg(outstr);};
  readerror = 1;
  return(0.0);
 };
}

REAL readchar()   // reads data in compressed format 
{
int ch2;
readerror = 0;
ch2 = readch();
do {
          switch (ch2) {
case '\n':
case ' ': ch2 = readch();
          break;
case '1': return(1.0);
case '0': return(0.0);
case '*': do ch2 = readch(); while(ch2 != '\n');
          break;
case EOF: readerror = 2;
          return(0.0);
default:  texterror(566);
          readerror = 1;
          return(0.0);};
} while (0 == 0);
}

void clearalist(NODE *thelist)
{
NODE *list, *lastnode;
REAL *vec;

list = thelist;
while (list != NULL)
 {
  vec = list->cvector;
  free(vec);
  lastnode = list;
  list = list->next;
  free(lastnode);
 };
thelist = NULL;
}

NODE *readpats(char *thefile,NODE *thelist)
{
NODE *list, *lastnode, *newnode;
REAL *vec, *newvector;
int j;
FILE *patterns;

if (!pushfile(thefile)) return(NULL);

// dispose of the old list, if there is one 

if (thelist != NULL) clearalist(thelist);

// set up an unused node at the beginning 

list = (NODE *) malloc(sizeof(NODE));
if (list == NULL)
   {pg("ran out of memory\n"); exit(420);};
lastnode = list;
lastnode->next = NULL;
listcount = 0;

// read in the patterns 
ch = readch();
while (ch != EOF)
 {
  if (ch != '*')
   {
    listcount = listcount + 1;
    newnode = (NODE *) malloc(sizeof(NODE));
    if (newnode == NULL)
       {pg("ran out of memory\n"); exit(420);};
    newvector = (REAL *) malloc(sizeof(REAL) * length);
    if (newvector == NULL)
       {pg("ran out of memory\n"); exit(420);};
    newnode->cvector = newvector;
    newnode->next = NULL;
    newnode->nn = 0;
    lastnode->next = newnode;
    bufferptr = bufferptr - 1;
    for (j=1;j<=length;j++)
     {
      *newvector++ = readreal(GE,-MAXVAL,'r');
      if (readerror)
       {
        sprintf(outstr,"read error, pattern %d\n",listcount);
        pg(outstr);
        goto endsbad;
       };
     };
    newnode->correctclass = readint(1,nclasses,'r');
    if (readerror) goto endsbad;
    lastnode = newnode;
   };
  do ch = readch(); while (ch != '\n');
  ch = readch();
 };
fclose(dataf);
popfile();
thelist = list->next;
free(list);
return(thelist);

endsbad:
fclose(dataf);
popfile();
if (list != NULL) clearalist(thelist);
return(NULL);
}

void saveweights(char *fn)
{
NODE *p;
REAL *vec;
int j;
FILE *f;

f = fopen(fn,"w");
if (f == NULL)
 {
  sprintf(outstr,"cannot open: %s\n",fn); pg(outstr);
  return;
 };
p = protolist;
while (p != NULL)
 {
  vec = p->cvector;
  for (j=1;j<=length;j++) fprintf(f," %f\n",*vec++);
  fprintf(f," %d\n",p->correctclass);
  p = p->next;
 };
fflush(f);
fclose(f);
}

void knn(NODE *known, NODE *unknown) // k nearest neighbor 
{
register REAL sum;
register REAL diff;
register REAL *uvec;
register REAL *pvec;
PAT *p, *dummy, *end, *c, *nextc;
int i, j, pointnumber;
short pattern,large,place,m,right,wrong,answer,rtanswer;
REAL pct, ipct;
INT32 *addr;
NODE *u, *protopoint, *pointptr;

right = 0;
wrong = 0;
addr = errmat;
for (i=0;i<=errmatsize-1;i++) *addr++ = 0;

// free any nodes that have already been used 

c = closest;
while (c != NULL) {nextc = c->next; free(c); c = nextc;};

// create the list that contains the closest values 

pg("working ...\n");
closest = (PAT *) malloc(sizeof(PAT));
if (closest == NULL)
   {pg("ran out of memory\n"); exit(420);};
closest->distsq = -1.0;
closest->point = NULL;
p = closest;
for (i=1;i<=k;i++)
 {
  c = (PAT *) malloc(sizeof(PAT));
  if (c == NULL)
     {pg("ran out of memory\n"); exit(420);};
  p->next = c;
  c->next = NULL;
  p = c;
 };
end = c;

// find the k closest neighbors 
pointnumber = 0;
u = unknown;
while (u != NULL)
 {
  c = closest->next;
  while (c != NULL)
   {
    c->distsq = MAXVAL;
    c->point = NULL;
    c = c->next;
   };
  protopoint = known;
  while (protopoint != NULL)
   {
    pvec = protopoint->cvector;
    uvec = u->cvector;
    if (pvec != uvec)
     {
      sum = 0.0;
      for (j=1;j<=length;j++)
       {
        diff = *uvec++ - *pvec++;
        sum = sum + diff * diff; 
       };
// if this point is closer than the farthest, move this pattern in 
// sprintf(outstr,"%f %f\n",sum,end->distsq); pg(outstr); 
      if (sum < end->distsq)
       {
        c = closest->next;
        while (c != NULL && sum > c->distsq) c = c->next;
        if (c == end)
         {
          c->distsq = sum;
          c->point = protopoint;
         }
        else
         {
          end->point = c->point;
          end->distsq = c->distsq;
          end->next = c->next;
          c->distsq = sum;
          c->next = end;
          c->point = protopoint;
          c = end->next;
          while (c->next != end) c = c->next;
          c->next = NULL;
          end = c;
         };
       }; // end if sum < 
     }; // end if uvec != pvec 
    protopoint = protopoint->next;
   }; // end while protopoint 
  if (k == 1)
   {
    pointptr = end->point;
    answer = pointptr->correctclass;
// sprintf(outstr,"%x %x %d\n",end,pointptr,answer); pg(outstr); 
   }
  else
   {
    addr = votes;
    for (m=0;m<=nclasses;m++) *addr++ = 0;
    c = closest->next;
    while (c != NULL)
     {
      pointptr = c->point;
      answer = pointptr->correctclass;
      *(votes + answer) = *(votes + answer) + 1;
      c = c->next;
     };
    large = *(votes + 1);
    place = 1;
    for (m=2;m<=nclasses;m++)
    if (*(votes + m) > large)
     {
      large = *(votes + m);
      place = m;
     };
    answer = place;
   }; // end else 
  rtanswer = u->correctclass;
  if (answer == rtanswer) right = right + 1; else wrong = wrong + 1;
  addr = errmat + (answer - 1) * (nclasses + 1) + rtanswer;
  *addr = *addr + 1;
  if (summary == '-')
   {
    uvec = u->cvector;
    pointnumber = pointnumber + 1;
    sprintf(outstr,"%5d",pointnumber);
    pg(outstr);
    for (j=1;j<=length;j++)
     {
      sprintf(outstr,"%7.3lf ",*uvec++);
      pg(outstr);
     };
    sprintf(outstr,"%5d ",answer);
    pg(outstr);
    if (answer == rtanswer) j = pg("ok\n"); else j = pg("\n");
    if (j) return;
   };
#ifndef UNIX
  if (kbhit() && getch() == 27) return;
#endif
  u = u->next;
 }; // end while u 
      

if (unknown == blank)
 {
  sprintf(outstr,"answer = %d\n",answer); pg(outstr);
  return;
 };
addr = errtotal;
for (i=0;i<=nclasses;i++) *addr++ = 0;
for (i=1;i<=nclasses;i++)
 {
  sum = 0;
  for (j=1;j<=nclasses;j++)
   {
    addr = errmat + (i - 1) * (nclasses + 1) + j;
    sum = sum + *addr;
   };
  *(errtotal + i) = sum;
 };
pg("   ");
for (i=1;i<=nclasses;i++) {sprintf(outstr,"%3d",i); pg(outstr);};
pg("\n");
for (i=1;i<=nclasses;i++)
 {
  sprintf(outstr,"%3d",i); pg(outstr);
  for (j=1;j<=nclasses;j++)
   {
    addr = errmat + (i - 1) * (nclasses + 1) + j;
    if (*(errtotal + i) == 0) pg("  *");
    else
     {
      ipct = (*addr * 100.0) / *(errtotal + i) + 0.5;
      sprintf(outstr,"%3.0f",ipct); pg(outstr);
     };
   };
  pg("\n");
 };
pg("   ");
for (i=1;i<=nclasses;i++) {sprintf(outstr,"%3d",i); pg(outstr);};
pg("\n\n");
pct = 100.0 * right / (right + wrong);
sprintf(outstr,"pct = %8.4f  right = %d  wrong = %d\n",pct,right,wrong);
pg(outstr);
}

void cluster(int theclass, NODE *list)
{
register REAL dist, sum, *cvec, *patvec;
register REAL min, old, maxdistsq;
register int i,j,classcount;
CPAT *cl, *prev, *newpat, *closest, *cluster, *dump;
NODE *pattern;

create = 0;
maxdistsq = maxdist * maxdist;
cl = clusterstart->next;
while (cl != NULL)
 {
  cl->cnt = 0;
  cl = cl->next;
 };
pattern = list;
classcount = 0;
while (pattern != NULL)
// if (pattern->correctclass != theclass || pattern->nn) 
if (pattern->correctclass != theclass) pattern = pattern->next;
else
 {
  classcount = classcount + 1;
  cluster = clusterstart->next;
  min = MAXVAL;
  while (cluster != NULL)
   {
    patvec = pattern->cvector;
    cvec = cluster->cvector;
    sum = 0.0;
    for (i=1;i<=length;i++)
     {
      dist = *patvec++ - *cvec++;
      sum = sum + dist * dist;
     };
    if (sum < min)
     {
      closest = cluster;
      min = sum;
     };
    cluster = cluster->next;
   };
  if (min > maxdistsq)
   {
    create = create + 1;
    count = count + 1;
    cl = clusterstart->next;
    newpat = (CPAT *) malloc(sizeof(CPAT));
    if (newpat == NULL)
       {pg("ran out of memory\n"); exit(420);};
    clusterstart->next = newpat;
    newpat->next = cl;
    newpat->cnt = 1;
    cvec = (REAL *) malloc(sizeof(REAL) * length);
    if (cvec == NULL)
       {pg("ran out of memory\n"); exit(420);};
    newpat->cvector = cvec;
    patvec = pattern->cvector;
    for (i=1;i<=length;i++) *cvec++ = *patvec++;
   }
  else
   {
    closest->cnt = closest->cnt + 1;
    cvec = closest->cvector;
    patvec = pattern->cvector;
    for (i=1;i<=length;i++)
     {
      old = *cvec;
      *cvec = oneminus * old + lambda * *patvec++;
      cvec++;
     };
   };
  pattern = pattern->next;
 };
destroy = 0;
prev = clusterstart;
cl = prev->next;
while (cl != NULL)
 if (cl->cnt == 0 || cl->cnt < (fract * classcount))
  {
   destroy = destroy + 1;
   count = count - 1;
   dump = cl;
   prev->next = cl->next;
   cl = cl->next;
   free(dump->cvector);
   free(dump);
  }
 else
  {
   prev = cl;
   cl = cl->next;
  };
}

void clusterall(NODE *list)
{
 int classcnt, pass, i;
 NODE *p, *dummy;
 CPAT *cl, *cnext;
 REAL *cvec, *patvec;

nprotos = 0;
dummy = (NODE *) malloc(sizeof(NODE));
if (dummy == NULL)
   {pg("ran out of memory\n"); exit(420);};
dummy->next = NULL;
if (clusterstart == NULL)
 {
  clusterstart = (CPAT *) malloc(sizeof(CPAT));
  if (clusterstart == NULL)
     {pg("ran out of memory\n"); exit(420);};
  clusterstart->next = NULL;
  cvec = (REAL *) malloc(sizeof(REAL) * length);
  if (clusterstart == NULL)
     {pg("ran out of memory\n"); exit(420);};
  for (i=1;i<=length;i++) *cvec++ = MAXVAL; 
 }; 
for (classcnt=1;classcnt<=nclasses;classcnt++)
 {
// printf("classcnt %d\n",classcnt); 
//#ifndef UNIX
//  interrupt(814);
//#endif
  cl = clusterstart->next;
  while (cl != NULL) {cnext = cl->next; free(cl); cl = cnext;};
  clusterstart->next = NULL;
  count = 0;
  maxpasses = 10;
  pass = 0;
  create = 1;
  destroy = 1;
  unstable = 1;
  lambda = initiallambda;
  while (pass <= maxpasses && (unstable || pass < 3))
   {
    oneminus = 1.0 - lambda;
    cluster(classcnt,list);
// printf("pass %d\n",pass); 
    unstable = create + destroy;
    pass = pass + 1;
    lambda = lambda / 2.0;
//#ifndef UNIX
//    interrupt(833);
//#endif
   };
  nprotos = nprotos + count;
  sprintf(outstr,"pattern %d has %d clusters\n",classcnt,count);
  pg(outstr);
// at this point copy over to prototypes 
  cl = clusterstart->next;
  while (cl != NULL)
   {
    p = (NODE *) malloc(sizeof(NODE));
    if (p == NULL)
       {pg("ran out of memory\n"); exit(420);};
    p->cvector = cl->cvector;
    p->correctclass = classcnt;
    p->next = dummy->next;
    dummy->next = p;
    cl = cl->next;
   };
 }; // end for class 
//#ifndef UNIX
//interrupt(849);
//#endif
protolist = dummy->next;
}

NODE *nn(NODE *trainpoint,int answer)  // finds the nearest prototype 
{
register REAL *protovec, *trainvec;
register REAL diff, sum;
register short i,j;
register REAL min;
NODE *protopoint, *place;

min = MAXVAL;
place = NULL;
  
protopoint = protolist;
while (protopoint != NULL)
 {
  if (answer == -1 || protopoint->correctclass == answer)
   {
    protovec = protopoint->cvector;
    trainvec = trainpoint->cvector;
    sum = 0.0;
    for (j=1;j<=length;j++) // find difference, square and sum 
     {
      diff = *trainvec++ - *protovec++;
      sum = sum + diff * diff;
     };
    if (sum < min)
     {
      min = sum;
      place = protopoint;
     };
   };
  protopoint = protopoint->next;
 };
return(place); // the pointer to the closest prototype 
}

// the learning vector quantization 1 algorithm 
void lvq1(NODE *point,NODE *nearestproto,int ans)
{
register REAL ralpha, *trainvec, *protovec;
register short j;

ralpha = alpha;
if (nearestproto->correctclass != ans)  // nearest neighbor is wrong 
 {
  trainvec = point->cvector;
  protovec = nearestproto->cvector;
  for (j=1;j<=length;j++)
   {
    *protovec = *protovec - ralpha * (*trainvec++ - *protovec);
    protovec++;
   }
 }
else // nearest neighbor is right 
 {
  trainvec = point->cvector;
  protovec = nearestproto->cvector;
  for (j=1;j<=length;j++)
   {
    *protovec = *protovec + ralpha * (*trainvec++ - *protovec);
    protovec++;
   };
 };
}

// decision surface mapping from March 1991 IEEE Trans on NN 

void dsm(NODE *trainpoint,NODE *nearestpoint,int ans)
{
register REAL *trainvec, *protovec;
register REAL ralpha;
register short j;
NODE *correct;

ralpha = alpha;

// wrong answers require a correction; right ones do not 

if (nearestpoint->correctclass != ans) // if wrong 
 { // move nearest prototype away 
  trainvec = trainpoint->cvector;
  protovec = nearestpoint->cvector;
  for (j=1;j<=length;j++)
   {
    *protovec = *protovec - ralpha * (*trainvec++ - *protovec);
    protovec++;
   };
 // find nearest prototype in the correct class 
  correct = nn(trainpoint,ans);
  trainvec = trainpoint->cvector;
  protovec = correct->cvector;
// move it closer 
  for (j=1;j<=length;j++)
   { 
    *protovec = *protovec + ralpha * (*trainvec++ - *protovec);
    protovec++;
   };
 };
}

int nonetwork()
{
if (length == 0 || nclasses == 0)
   {pg("there is no network.\n"); return(1);}
else return(0);
}

int nopatterns(int list)
{
if (list == TRAIN && trainlist != NULL) return(0);
else if (list == TEST && testlist != NULL) return(0);
else {pg("there are no patterns\n"); return(1);};
}

int noprotos()
{if (protolist == NULL) return(1); else return(0);}

void evalone(int list, int thepat)
{
 NODE *point, *nearest;
 int i, answer, rtanswer;
 REAL *coord;

 sprintf(outstr,"%5d ",thepat);
 pg(outstr);
 if (list == TRAIN) point = trainlist; point = testlist;
 for (i=1;i<=(thepat-1);i++) point = point->next;
 nearest = nn(point,-1);
 coord = point->cvector;
 for (i=1;i<=length;i++)
  {
   sprintf(outstr,"%8.4f ",*coord++);
   pg(outstr);
  };
 answer = nearest->correctclass;
 rtanswer = point->correctclass;
 sprintf(outstr,"%5d ",answer);
 pg(outstr);
 if (answer == rtanswer)
  {
   if (pg("ok\n")) return;
  }
 else if (pg("\n")) return;
}

void eval(int list,int  printing)
{
NODE *nearestproto, *point;
int answer, rtanswer, right, wrong, i, patcount, stop;
REAL pct, *coord;

if (noprotos()) return;
patcount = 0;
right = 0;
wrong = 0;
for (i=0;i<=nclasses;i++) *(errtotal + i) = 0;
for (i=0;i<=errmatsize-1;i++) *(errmat + i) = 0;
if (list == TRAIN) point = trainlist; else point = testlist;
while (point != NULL)
 {
  rtanswer = point->correctclass;
  nearestproto = nn(point,-1);
  answer = nearestproto->correctclass;
  point->nn = answer;
  if (answer == rtanswer) right = right + 1;
  else wrong = wrong + 1;
  if (printing)
   {
    patcount = patcount + 1;
    sprintf(outstr,"%5d ",patcount);
    pg(outstr);
    coord = point->cvector;
    if (listpat == '+')
     {
      for (i=1;i<=length;i++)
       {
        sprintf(outstr,"%8.4f ",*coord++);
        pg(outstr);
       };
     };
    sprintf(outstr,"%5d ",answer);
    pg(outstr);
    if (answer == rtanswer)
     {
      stop = pg("ok\n");
      if (stop == 1) return;
     }
    else
     {
      stop = pg("\n");
      if (stop == 1) return;
     };
   };
  point = point->next;
 };
pct = 100.0 * right / (right + wrong);
sprintf(outstr,"pct = %6.2f  right = %d  wrong = %d\n",pct,right,wrong);
pg(outstr);
s[list].iterno = s[TRAIN].iterno;
s[list].right = right;
s[list].wrong = wrong;
s[list].pctright = pct;
}

void makecopy() // of the original prototypes 
{
NODE *list, *lastnode, *newnode, *orig;
REAL *vec, *newvector, *origvector;
int j;
FILE *patterns;

// get rid of any old prototypes 

if (initialprotos != NULL)
 {
  list = initialprotos;
  while (list != NULL)
   {
    newnode = list->next;
    vec = list->cvector;
    free(vec);
    free(list);
    list = newnode;
   };
 };

// set up an unused node at the beginning 

initialprotos = (NODE *) malloc(sizeof(NODE));
if (initialprotos == NULL)
   {pg("ran out of memory\n"); exit(420);};
lastnode = initialprotos;

// copy the original set of prototypes 

orig = protolist;
while (orig != NULL)
 {
  newnode = (NODE *) malloc(sizeof(NODE));
  if (newnode == NULL)
     {pg("ran out of memory\n"); exit(420);};
  newvector = (REAL *) malloc(sizeof(REAL) * length);
  if (newvector == NULL)
     {pg("ran out of memory\n"); exit(420);};
  newnode->cvector = newvector;
  newnode->next = NULL;
  newnode->nn = 0;
  lastnode->next = newnode;
  origvector = orig->cvector;
  for (j=1;j<=length;j++) *newvector++ = *origvector++;
  newnode->correctclass = orig->correctclass;
  lastnode = newnode;
  orig = orig->next;
 };
initialprotos = initialprotos->next;
}

void run(int passes)
{
NODE *nearestproto, *point;
int i,j,p,answer;
REAL decrease, changes;

alpha = highalpha;
changes = passes * npats;
decrease = (highalpha - lowalpha) / changes;
outofdate = 1;

pg("running ...\n");
for (p=1;p<=passes;p++)
 {
  point = trainlist;
  while (point != NULL)
   {
    answer = point->correctclass;
    nearestproto = nn(point,-1);
    if (algorithm == 'd') dsm(point,nearestproto,answer);
    else lvq1(point,nearestproto,answer);
    point = point->next;
    alpha = alpha - decrease;
   };
  s[TRAIN].iterno = s[TRAIN].iterno + 1;
  eval(TRAIN,(int) 0);
#ifndef UNIX
  if (kbhit() && getch() == 27) return;
#endif
 }
}
 
void printmatrix(int list)
{int i, j, p, rtanswer, answer;
 NODE *point;
 REAL pct, ipct;
 INT32 *addr;
if (list == TRAIN) point = trainlist; else point = testlist;
if (s[list].iterno == 0) eval(list,(int) 0);
else if (list == TEST && s[list].iterno != s[TRAIN].iterno)
   eval(list,(int) 0);
while (point != NULL)
 {
  rtanswer = point->correctclass;
  answer = point->nn;
  addr = errmat + (answer - 1) * (nclasses + 1) + rtanswer;
  *addr = *addr + 1;
  *(errtotal + answer) = *(errtotal + answer) + 1;
  point = point->next;
 };
// print the error matrix; when the program claims class i (row i)
// is the answer, what percentage of the time is it really class j
// (column j) 
pg("   ");
for (i=1;i<=nclasses;i++) {sprintf(outstr,"%3d",i); pg(outstr);};
pg("\n");
for (i=1;i<=nclasses;i++)
 {
  sprintf(outstr,"%3d",i); pg(outstr);
  for (j=1;j<=nclasses;j++)
   {
    addr = errmat + (i - 1) * (nclasses + 1) + j;
    if (*(errtotal + i) == 0) pg("  *");
    else
     {
      ipct = (*addr * 100.0) / *(errtotal + i) + 0.5;
      sprintf(outstr,"%3.0f",ipct); pg(outstr);
     };
   };
  pg("\n");
 };
pg("   ");
for (i=1;i<=nclasses;i++) {sprintf(outstr,"%3d",i); pg(outstr);};
pg("\n\n");
pct = 100.0 * right / (right + wrong);
sprintf(outstr,"pct = %8.4f  right = %d  wrong = %d\n",pct,right,wrong);
}

void help()
{
int ch2;
putchar('\n');
do ch2 = readch(); while (ch2 == ' ' && ch2 != '\n');
         switch (ch2) {
default:
pg("For help type h followed by the letter of the command.\n");
if (ch2 == '\n') bufferptr = bufferptr - 1;
break;

case '?':
pg("? prints program status and parameters.\n");
break;

case '*':
pg("* at the beginning of a line makes the line a comment.\n");
break;

case '!':
pg("Enter system command after the !\n");
break;

case 'A':
pg("A l sets the algorthm to lvq 1.\n");
pg("A d sets the algorithm to dsm.\n");
break;

case 'a':
pg("a <real> sets alpha to <real>.\n");
break;

case 'C':
pg("C <real1> <real2> <real3> uses the training set to form prototypes using\n");
pg("        a simple clustering algorithm with lambda = <real1>,\n");
pg("        <real1> / 2 and <real1> / 4 with maxdist = <real2>.\n");
pg("        Clusters with less than <real3> * number of patterns are deleted.\n");
break;

case 'c':
pg("c (clear) resets the prototypes and the other parameters.\n");
break;

case 'd':
pg("d <real> does one pass of DSM with alpha = <real>.\n");
pg("d <real1> <real2> <int> does <int> passes of dsm starting with\n");
pg("        alpha = <real1> and ending with alpha = <real2>.\n");
break;

case 'f':
pg("f c+ makes a copy of all I/O on the file, copy.\n");
pg("f c- stops making a copy of all I/O.\n");
pg("f e+ echos input.\nf e- stops echoing input.\n");
pg("f P <int> sets page size to <int>.\n");
pg("f P 0 stops paging.\n");
pg("f s+ to summarize training and testing results (knn only).\n");
pg("f s- to list all training and testing patterns (knn only).\n");
pg("f w+ writes patterns when evaluating them with p and t\n");
pg("f w- does not write patterns when evaluating them with p and t\n");
break;

case 'h':
pg("h <char> gives help with command, <char>.\n");
break;

case 'i':
pg("i <filename> takes commands from <filename>.\n");
break;

case 'l':
pg("l <real> does one pass of lvq1 with alpha = <real>.\n");
pg("l <real1> <real2> <int> does <int> passes of lvq1 starting with\n");
pg("        alpha = <real1>\n   and ending with alpha = <real2>.\n");
break;

case 'm':
pg("m <int1> <int2> ""makes a network"" with <int1> input units (the\n");
pg("        length of the pattern) and <int2> output units (the\n");
pg("        number of classes.\n");
break;

case 'n':
pg("<int> n  or <int> nn classifies training set patterns using the\n");
pg("        <int> nearest neighbors using the training set as\n");
pg("        prototypes but excluding the training set point itself.\n");
pg("<int> n <pattern>  or  <int> nn <pattern>\n");
pg("        classifies the <pattern> using the <int> nearest neighbors.\n");
pg("<int> n <input pattern>  >  <file>\n");
pg("<int> nn <input pattern>  >  <file> writes the <int> nearest training\n");
pg("        set vectors to the <file> with the distance written\n");
pg("        out preceeded by an *.  If no file is given the last\n");
pg("        file named is used or if no file has been named, the\n");
pg("        file, nearest is used.\n");
break;

case 'p':
pg("p prints training patterns and their answers.\n");
pg("p0 evaluates the training patterns and summarizes the results.\n");
pg("p <int> evaluates training pattern <int>.\n");
pg("pm prints the confusion matrix.\n");
break;

case 'P':
pg("P <pattern> classifies the pattern using its nearest prototype\n");
break;

case 'q':
pg("q quits the program.\n");
break;

case 'r':
pg("rp <filename> reads prototype vectors (""weights"").\n\n");
pg("rt <filename> reads training set vectors from <filename>.\n\n");
pg("rw <filename> reads prototype vectors (""weights"").\n\n");
pg("r <real> runs one pass through the training set using alpha =");
pg(" <real>.\n");
pg("r <real1> <real2> <int> runs <int> passes through the training ");
pg("set starting\n   with alpha = <real1> and ending with alpha = ");
pg("<real2>.\n");
break;

case 's':
pg("sp saves prototypes to the current prototype file.\n");
pg("sp <filename> saves prototypes to the file <filename>.\n");
pg("sw saves prototypes (""weights"") to the current prototype file.\n");
pg("sw <filename> saves prototypes to the file <filename>.\n");
break;

case 't':
pg("t evaluates and prints all the test patterns.\n");
pg("t 0 evaluates test patterns and summarizes the results.\n");
pg("t <int> evaluates test pattern <int>.\n");
pg("t m prints the confusion matrix for the test patterns.\n");
pg("tf <filename> reads the test patterns on <filename>.\n");
break;
         }
pg("\n");
}

void scanrun()
{ REAL rtemp;
  int itemp;
if (protolist == NULL) {pg("there are no prototypes\n"); return;};
if (trainlist == NULL)
   {pg("there are no training patterns\n"); return;};
if (ch == '\n' || ch == '*') {npasses = 1; run(npasses); return;}
else
 {
  bufferptr = bufferptr - 1;
  rtemp = readreal(GE,0.0,'r');
  if (readerror) return; else highalpha = rtemp;
  do ch = readch(); while (ch == ' ');
  bufferptr = bufferptr - 1;
  if (ch == '\n')
   {
    lowalpha = highalpha;
    npasses = 1;
    run(npasses);
    return;
   };
  rtemp = readreal(GE,0.0,'r');
  if (readerror) return; else lowalpha = rtemp;
  itemp = readint(1,MAXINT,'r');
  if (readerror) return;
  npasses = itemp;
  run(npasses);
 };
} // end of scanrun 

void restartcmdloop(int dummy)
{
int popresult;
popresult = OK;
while (dataf != stdin && popresult == OK) popresult = popfile();
signal(SIGINT,restartcmdloop);
longjmp(cmdloopstate,1);
}
 
void cmdloop()    // read commands and process them 
{
int finished;
int i, itemp, itemp2;
char string[81];
REAL rtemp, *vec;
NODE *point, *answeraddr;

setjmp(cmdloopstate);  // restart here from SIGINT 
finished = 0;
ch = ' ';
do{
   if (dataf == stdin) pg("[?!*AaCcdfhlmnPpqrst]? ");
   if (dataf == stdin) lineno = 0;
   while(ch == ' ' || ch == '\n') ch = readch();
   switch (ch) {

case EOF:
if (dataf == stdin)
 {
  pg("taking commands from stdin now\n");
  ch = ' ';
 }
else popfile();
break;

case '?':
sprintf(outstr,"\n%d iterations %d prototypes\n",s[TRAIN].iterno,nprotos);
pg(outstr);
sprintf(outstr,"data file = %s\n",datafilename);
pg(outstr);
sprintf(outstr,"input file = %s\n",inputfile);
pg(outstr);
sprintf(outstr,"nearest neighbors file = %s\n",nearestfile);
pg(outstr);
sprintf(outstr,"original prototype file = %s\n",original);
pg(outstr);
sprintf(outstr,"new prototype file = %s\n",newproto);
pg(outstr);
if (algorithm == 'l') pg("lvq1 "); else pg("dsm ");
sprintf(outstr,"%7.5f %7.5f %d\n",highalpha,lowalpha,npasses);
pg(outstr);
sprintf(outstr,"f c%c e%c P %1d s%c w%c\n",copyflag,echo,pagesize,summary,listpat);
pg(outstr);
sprintf(outstr,"%d training patterns from the file %s\n",npats,trainfile);
pg(outstr);
sprintf(outstr,"%3d iterations ",s[TRAIN].iterno); pg(outstr);
sprintf(outstr," %5.2f%%  right ",s[TRAIN].pctright); pg(outstr);
sprintf(outstr,"(%d right,  %d wrong)\n",s[TRAIN].right,s[TRAIN].wrong);
pg(outstr);
sprintf(outstr,"%d testing patterns from the file %s\n",ntestpats,testfile);
pg(outstr);
sprintf(outstr,"%3d iterations ",s[TEST].iterno); pg(outstr);
sprintf(outstr," %5.2f%%  right ",s[TEST].pctright); pg(outstr);
sprintf(outstr,"(%d right,  %d wrong)\n",s[TEST].right,s[TEST].wrong);
pg(outstr);
pg("for help, type h followed by the letter of the command\n\n");
break;

case '!':
i = 0;
ch = readch();
while (ch != '\n' && i <= 80)
 {
  string[i] = ch;
  ch = readch();
  i = i + 1;
 };
bufferptr = bufferptr - 1;
string[i] = '\0';
system(string);
break;

case '*': break;  // * on a line is a comment 

case 'A':
do ch = readch(); while (ch == ' ' || ch == '\n');
if (ch == 'l' || ch == 'd') algorithm = ch;
break;

case 'a':
rtemp = readreal(GE,0,'a');
if (readerror) break;
alpha = rtemp;
break;

case 'c':
if (nopatterns(TRAIN)) goto endc;
point = protolist;
while (point != NULL) {point->nn = 0; point = point->next;};
point = trainlist;
while (point != NULL) {point->nn = 0; point = point->next;};
point = testlist;
while (point != NULL) {point->nn = 0; point = point->next;};
for (i=TRAIN;i<=TEST;i++)
 {
  s[i].right = 0;
  s[i].wrong = 0;
  s[i].pctright = 0;
  s[i].iterno = 0;
 };
{NODE *orig, *newnode;
 REAL *origvector, *newvector;
 int j;
 orig = initialprotos;
 newnode = protolist;
 while (orig != NULL)
  {
   origvector = orig->cvector;
   newvector = newnode->cvector;
   for (j=1;j<=length;j++) *newvector++ = *origvector++;
   newnode = newnode->next;
   orig = orig->next;
  };
 };
endc: break;

case 'C':
if (nonetwork() || nopatterns(TRAIN)) goto endC;
initiallambda = readreal(GE,0.0,'C');
if (readerror) goto endC;
maxdist = readreal(GT,0.0,'C');
if (readerror) goto endC;
do ch = readch(); while (ch == ' ');
if (ch >= '0' && ch <= '9')
 {
  bufferptr = bufferptr - 1;
  fract = readreal(GE,0.0,'C');
  if (readerror) goto endC;
 }
else fract = 0.0;
clusterall(trainlist);
s[TRAIN].iterno = 0;
s[TEST].iterno = 0;
for (i=TRAIN;i<=TEST;i++)
 {
  s[i].right = 0;
  s[i].wrong = 0;
  s[i].pctright = 0;
  s[i].iterno = 0;
 };
makecopy();
endC: break;

case 'h': help(); break;

case 'f':
while (ch != '\n' && ch != '*')
{
 ch = readch();
 if (ch == 'c')
  {
   do ch = readch(); while (ch == ' ');
   if (ch == '+' || ch == '-')
    {
     copyflag = ch;
     if (copyflag == '+') copy = fopen("copy","w");
     else if (copy != NULL)
      {
       fflush(copy);
       fclose(copy);
      }
     else pg("the copy file was never opened\n");
    }
   else texterror(1697);
  }
 else if (ch == 'e')
  {
   do ch = readch(); while (ch == ' ');
   if (ch == '+' || ch == '-') echo = ch;
   else texterror(1703);
  }
 else if (ch == 'P')
  {
   itemp = readint(0,MAXINT,'f');
   if (!readerror) pagesize = itemp;
  }
 else if (ch == 's')
  {
   do ch = readch(); while (ch == ' ');
   if (ch == '+' || ch == '-') summary = ch;
  }
 else if (ch == 'w')
  {
   do ch = readch(); while (ch == ' ');
   if (ch == '+' || ch == '-') listpat = ch;
  }
 else if (ch == ' ' || ch == '*' || ch == '\n') /* nothing */ ;
 else texterror(1721);
};
bufferptr = bufferptr - 1;
break;

case 'i':
do ch = readch(); while (ch == ' ');
if (ch != '\n')
{
 bufferptr = bufferptr - 1;
 inputfile = readstr();
}
pushfile(inputfile);
break;

case 'm':
length = readint(0,MAXINT,'m');
nclasses = readint(0,MAXINT,'m');
errmatsize = nclasses * (nclasses + 1);
errmat = (INT32 *) malloc(sizeof(INT32) * nclasses * (nclasses + 1));
if (errmat == NULL)
   {pg("ran out of memory\n"); exit(420);};
errtotal = (INT32 *) malloc(sizeof(INT32) * (nclasses + 1));
if (errtotal == NULL)
   {pg("ran out of memory\n"); exit(420);};
votes = (INT32 *) malloc(sizeof(INT32) * (nclasses + 1));
if (votes == NULL)
   {pg("ran out of memory\n"); exit(420);};
blank = (NODE *) malloc(sizeof(NODE));
if (blank == NULL)
   {pg("ran out of memory\n"); exit(420);};
vec = (REAL *) malloc(sizeof(REAL) * (length + 1));
if (vec == NULL)
   {pg("ran out of memory\n"); exit(420);};
blank->cvector = vec;
blank->next = NULL;
break;

case '1':
case '2':
case '3':
case '4':
case '5':
case '6':
case '7':
case '8':
case '9': // case 'n' 
if (nonetwork() || nopatterns(TRAIN)) goto endn;
bufferptr = bufferptr - 1;
k = readint(1,npats,'n');
if (readerror) goto endn;
do ch = readch(); while (ch == ' ');
if (ch != 'n') texterror(1763);
else
 {
  do ch = readch(); while (ch == ' ');
  if (ch == 'n') do ch = readch(); while (ch == ' ');
  if (ch == '\n' || ch == '*') knn(trainlist,trainlist);
  else
   {
    bufferptr = bufferptr - 1;
    vec = blank->cvector;
    for (i=1;i<=length;i++)
     {
      rtemp = readreal(GT,-MAXVAL,'n');
      if (readerror) goto endn;
      *vec++ = rtemp;
     };
    blank->correctclass = 0;
    do ch = readch(); while (ch == ' ');
    if (ch == '\n' || ch == '*') knn(trainlist,blank);
    else if (ch == '>')
     {PAT *c; NODE *point; REAL *coord; FILE *nearest;
      do ch = readch(); while (ch == ' ');
      if (ch != '\n' && ch != '*')
       {
        bufferptr = bufferptr - 1;
        nearestfile = readstr();
       };
      knn(trainlist,blank);
      c = closest->next;
      nearest = fopen(nearestfile,"w");
      coord = blank->cvector;
      fprintf(nearest,"* ");
      for (i=1;i<=length;i++) fprintf(nearest," %6.3f",*coord++);
      fprintf(nearest," *  0\n");
      while (c != NULL)
       {
        point = c->point;
        coord = point->cvector;
        for (i=1;i<=length;i++) fprintf(nearest," %6.3f",*coord++);
        fprintf(nearest,"%3d * %6.3f\n",point->correctclass,sqrt(c->distsq));
        c = c->next;
       };
      fflush(nearest);
      fclose(nearest);
     };
   };
 };
endn: break;

case 'p':
if (nonetwork() || nopatterns(TRAIN) || noprotos()) goto endp;
do ch = readch(); while (ch == ' ');
if (ch == '0') eval(TRAIN,(int) 0);
else if (ch == 'm') printmatrix(TRAIN);
else if (ch == '\n' || ch == '*') eval(TRAIN,(int) 1);
else if (ch >= '1' && ch <= '9')
 {
  bufferptr = bufferptr - 1;
  itemp = readint(1,npats,'p');
  if (!readerror) evalone(TRAIN,itemp);
 };
endp: break;

case 'P':
if (nonetwork() || noprotos()) goto endP;
vec = blank->cvector;
for (i=1;i<=length;i++)
 {
  rtemp = readreal(GT,-MAXVAL,'P');
  if (readerror) goto endP;
  *vec++ = rtemp;
 };
  answeraddr = nn(blank,-1);
  sprintf(outstr,"answer = %d\n",answeraddr->correctclass); pg(outstr);
endP: break;

case 'q': return;

case 'd':
algorithm = 'd';
do ch = readch(); while (ch == ' ');
scanrun();
break;

case 'l':
algorithm = 'l';
do ch = readch(); while (ch == ' ');
scanrun();
break;

case 'r':
if (nonetwork()) goto endr;
do ch = readch(); while (ch == ' ');
if (ch == 'p' || ch == 'w')
 {
  do ch = readch(); while (ch == ' ');
  bufferptr = bufferptr - 1;
  original = readstr();
  protolist = readpats(original,protolist);
  if (noprotos()) goto endr;
  nprotos = listcount;
  for (i=TRAIN;i<=TEST;i++)
   {
    s[i].iterno = 0;
    s[i].right = 0;
    s[i].wrong = 0;
    s[i].pctright = 0;
   };
  makecopy();
 }
else if (ch == 't')
 {
  do ch = readch(); while (ch == ' ');
  bufferptr = bufferptr - 1;
  trainfile = readstr();
  trainlist = readpats(trainfile,trainlist);
  if (nopatterns(TRAIN)) goto endr;
  npats = listcount;
  s[TRAIN].iterno = 0;
  s[TRAIN].right = 0;
  s[TRAIN].wrong = 0;
  s[TRAIN].pctright = 0.0;
  s[TRAIN].npats = npats;
 }
else scanrun();
endr: break;

case 's': // sp <filename> for save prototypes 
do ch = readch(); while (ch == ' ');
if (ch == 'p' || ch == 'w')
{
 do ch = readch(); while (ch == ' ');
 bufferptr = bufferptr - 1;
 if (ch == '*' || ch == '\n') /* nothing */ ; else newproto = readstr();
 if (nonetwork() || noprotos()) break; else saveweights(newproto);
}
else texterror(1899);
break;

case 't':
if (nonetwork()) break;
do ch = readch(); while (ch == ' ');
if (ch == 'f')
 {
  do ch = readch(); while (ch == ' ');
  bufferptr = bufferptr - 1;
  testfile = readstr();
  testlist = readpats(testfile,testlist);
  ntestpats = listcount;
  s[TEST].right = 0;
  s[TEST].wrong = 0;
  s[TEST].pctright = 0.0;
  s[TEST].npats = ntestpats;
  goto endt;
 }
if (nonetwork() || noprotos() || nopatterns(TEST)) goto endt;
if (ch == '\n' || ch == '*') eval(TEST,(int) 1);
else if (ch == 'm') printmatrix(TEST);
else if (ch == '0') eval(TEST,(int) 0);
else if (ch >= '1' && ch <= '9')
 {
  bufferptr = bufferptr - 1;
  itemp = readint(1,npats,'t');
  if (!readerror) evalone(TEST,itemp);
 }
else texterror(1928);
outofdate = 0;
endt: break;

default: texterror(1932);
break;
      };
    if (ch != '\n') do ch = readch(); while (ch != '\n');
  }while (!finished);
}

void main(int argc,char *argv[])
{
FILE *tempfile;
setbuf(stdout,NULL);  // set unbuffered output 
lineno = 0;
bufferend = 0;
bufferptr = BUFFSIZE + 1;
clusterstart = NULL;
closest = NULL;
copyflag = '-';
dataf = stdin;
// debug = fopen("debug","w"); 
echo = '-';
emptystring = '\0';
filestack[0] = stdin;
filestackptr = 0;
initialprotos = NULL;
nearestfile = "nearest";
nprotos = 0;
pagesize = 24;
listpat = '+';
protolist = NULL;
summary = '+';
if (argc == 1)
 {
  pg("no data file, stdin assumed\n");
  datafilename = &emptystring;
  testfile = &emptystring;
 }
else
 {
  datafilename = argv[1];
  pushfile(datafilename);
  if (argc == 3)
   {
    tempfile = fopen(argv[2],"r");
    if (tempfile == NULL)
     {
      sprintf(outstr,"there is no file: %s\n",argv[2]);
      pg(outstr);
      testfile = &emptystring;
     }
    else
     {
      testfile = argv[2];
      fclose(tempfile);
     }
   }
  else testfile = &emptystring;
 };
#if defined(UNIX) && defined(HOTKEYS)
initraw();
#endif
signal(SIGINT,restartcmdloop); // restart from interrrupt 
cmdloop();
if (debug != NULL) fclose(debug);
if (copy != NULL) fclose(copy);
}
